#!/usr/bin/python
## html.py -- module for generating nicely-indented HTML
## Author: Neale Pickett <neale@lanl.gov>
## Time-stamp: <99/01/26 10:00:46 neale>

## This software and ancillary information (herein called "SOFTWARE")
## called html.py made avaiable under the terms described here.  The
## SOFTWARE has been approved for release with associated LA-CC Number
## 89-47.

## Unless otherwise indicated, this SOFTWARE has been authored by an
## employee or employees of the University of California, operator of
## the Los Alamos National Laboratory under contract No. W-7405-ENG-36
## with the U.S. Department of Energy.  The U.S. Government has rights
## to use, reproduce, and distribute this SOFTWARE.  The public may
## copy, distribute, prepare derivative works and publicly display this
## SOFTWARE without charge, provided that this Notice and any statement
## of authorship are reproduced on all copies.  Neither the Government
## nor the University makes any warranty, express or implied, or assumes
## any liability or responsibility for the use of this SOFTWARE.

## If SOFTWARE is modified to produce derivative works, such modified
## SOFTWARE should be clearly marked, so as not to confuse it with the
## version available from LANL.

"""html.py -- module for generating nicely-indented HTML

Yeah, yeah, I realize that someone has already written HTMLgen, but it
doesn't do what I want it to do and (to me) it's too confusing.  Also,
it has this annoying habit of not letting you use arbitrary tags, and I
seem to recall that I needed an arbitrary tag, like 'align' or
something.  Anyway.

This module provides a set of classes you can use to create very
nice-looking HTML (indented and everything).  It's very simple to use
(IMHO), you just instantiate whatever class you want and keep calling
.append() on that class with data.  Here's an example:

  doc = html.document('Document Title', bgcolor='white')
  doc.par('This is the first paragraph.',
          'This is the second paragraph.')

  # Create a new paragraph object.  Any named variables we pass in will
  # show up as attributes in the resulting HTML tag
  p = html.par(align='center')
  p.append('This is a centered paragraph.')
  p.append(html.strong('Here is some strong text.'))

  # Have to remember to append the paragraph to the document
  doc.append(p)

There ya go.  I'm not going to document this any more until someone lets
me know they're using it.  :-p

Neale Pickett <neale@lanl.gov>

"""
# ' <- emacs font-lock fix :)

import sys, string, types, os, urllib

Referral = 'Referral'
indentation = '  '


def debug(msg):
    """Print out a page with a message.

    I was typing this in so much, I thought I may as well make it a
    function.

    """

    d = document("debugging message", msg)
    print_doc(d)
    

def get_cookies():
   """Returns a dict containing all the cookies the client sent.

   If it sent two or more cookies with the same name, it returns those
   in a list."""

   try:
      cookie_str = os.environ['HTTP_COOKIE']
   except KeyError:
      return {}

   cookies = {}

   for s in string.split(cookie_str, '; '):
      keyval = string.split(s, '=')
      try:
	 if type(cookies[keyval[0]]) == types.StringType:
	    cookies[keyval[0]] = [cookies[keyval[0]], keyval[1]]
	 else:
	    cookies[keyval[0]].append(keyval[1])
      except KeyError:
         if len(keyval) > 1:
            cookies[keyval[0]] = keyval[1]
         else:
            # Q: What kind of screwed up browser sends a cookie with no
            # data?
            #
            # A: MSIE, of course!
            pass

   return cookies


def print_doc(doc, headers=[], content_type='text/html'):
   """Prints out a document with headers.

   This is intended to be used with a document() or derived class as the
   first argument, but you can pass it anything with a string
   representation, or a list (in which case the list will be printed,
   one item per line).  The headers are a list and will be sent out
   exactly as they appear in the list (one list item per line).  Nothing
   will be prepended to them, so you have to format them yourself (eg
   "Cookies: feh")."""
   
   if type(doc) == types.ListType:
      s = string.join(doc, '\n')
   else:
      s = str(doc)

   print 'Content-Type: %s' % content_type
   print 'Content-Length: %d' % len(s)

   for i in headers:
      if (i != ''):
         print i

   print # end headers
   sys.stdout.write(s)

   sys.exit(0)


def redirect(url, headers=[]):
    """Redirect the browser to a new page.

    This does the redirection in the HTTP headers and *not* with a meta
    tag.  Which is the way it should be done, IMHO.  It looks just like
    the page Apache generates to redirect someone.

    """
    
    print 'Location: ' + url
    print 'Content-type: text/html'
    for i in headers:
      if (i != ''):
	 print i
    print

    #doc = document('Document moved')
    #doc.h(1, 'Document moved')
    #doc.par(literal('The document has moved',
    #                `href(url, 'here')` + '.'))
    #print doc
    #print url
    sys.exit(0)


def whoami(full_URI=0):
   """Finds my URI.

   Using environment variables, this function finds out what URI is
   being called for.  This is very useful for CGI scripts that need to
   refer to themselves.

   If full_URI is one, it will prepend an http(s)://."""

   if full_URI:
      try:
	 host = os.environ['HTTP_HOST']
      except KeyError:
	 import socket
	 host = socket.gethostbyname(socket.gethostname())
      try:
	 port = os.environ['SERVER_PORT']
      except KeyError:
	 port = "80"
      if os.environ.has_key('HTTPS'):
	 if port != '443':
	    URI = 'https://%s:%s' % (host, PORT)
	 else:
	    URI = 'https://%s' % host
      else:
	 if port != '80':
	    URI = 'http://%s:%s' % (host, PORT)
	 else:
	    URI = 'http://%s' % host
   else:
      URI = ''

   try:
      URI = URI + os.environ['SCRIPT_NAME']
   except KeyError:
      raise Referral, 'no referral information was provided by the HTTP server'
   try:
      URI = URI + os.environ['PATH_INFO']
   except KeyError:
      pass
   try:
      if os.environ['QUERY_STRING']:
	 URI = URI + '?' + os.environ['QUERY_STRING']
   except KeyError:
      pass

   return URI



def exception(type=None, value=None, tb=None, limit=None):
   """Returns a PRE element containing an exception.

   Lifted almost verbatim out of Guido's CGI module."""

   if type is None:
      type, value, tb = sys.exc_type, sys.exc_value, sys.exc_traceback

   import traceback

   e = escape()

   e.append(par(strong("Traceback (innermost last):")))
   e.append(pre(string.join(traceback.format_tb(tb, limit), "")))
   e.append(par(samp(strong(traceback.format_exception_only(type, value)[0]))))

   return e



def fields(fields=None, as_strings = 0):
   """Reads in form fields like the cgi module.

   This works pretty much like the cgi module, except that it returns a
   real dict, and all values in the dict are a list.  I was quite
   annoyed when I tried to pass in the same field name twice, only to
   find that what I thought was going to be a string was now a list.

   If you pass an existing FieldStorage as an argument, you will get
   back a dict representation of that object."""

   if not fields:
      import cgi
      fields = cgi.FieldStorage()

   out = {}
   for i in fields.keys():
      v = fields[i].value
      if type(v) == types.ListType:
         if as_strings:
            raise TypeError, "Multiple values for %s" % i
         else:
	    out[i] = v
      else:
         if as_strings:
	    out[i] = v
         else:
	    out[i] = [v]

   return out



def subst_file(filename, **vars):
   file = open(filename, 'r')
   lines = map(lambda l: l[:-1], file.readlines())
   file.close()

   return apply(subst_lines, (lines,), vars)



def subst_lines(lines, **vars):

   import regex

   container = regex.compile('\(<!-- \)?##\([-_A-Za-z0-9]+\)##\( -->\)?')

   outlines = []
   for line in lines:
      while container.search(line) != -1:
	 try:
	    replacement = str(vars[container.group(2)])
	 except KeyError:
	    replacement = str(comment('Unmached variable: ' + container.group(2)))

	 pre = line[:container.regs[0][0]]
	 post = line[container.regs[0][1]:]

	 if string.strip(pre) == '':
	    # pre is just whitespace, so pad our replacement's lines with that space
	    lines = string.split(replacement, '\n')
	    new = [lines[0]]
	    for l in lines[1:]:
	       new.append(pre + l)
	    replacement = string.join(new, '\n')

	 line = "%s%s%s" % (pre, replacement, post)

      outlines.append(line)

   return outlines


class element:
   """Base element class.

   This class defines a few useful methods which can be overridden.
   Probably the ones you'll want to look at are get_head and get_foot.
   It may also suffice to simply overload __init__ to call
   container.__init__ and then set self.head.

   One kinda nifty thing is that when you override this, the default
   head and foot elements will be the name of the new class.  Look down
   later in this file to see how that's done."""


   indent_level = 0
   head = None
   prefix_string = ""
   postfix_string = ""
   
      
   def __init__(self, **tags):
      """Extract useful stuff."""

      # Oh, wow, this is like, so cool.
      if self.head is None:
	 self.head = string.upper(self.__class__.__name__)

      self.tags = {}

      for tag, val in tags.items():
         self.set_tag(tag, val)

   def __repr__(self):
      return string.join(self.list(), '\n')


   def list(self, indent_level=0):
      lines = []
      indent = indentation * indent_level
      
      head = self.get_head()
      if head:
	 lines = _indent(head, indent_level)
	 
      foot = self.get_foot()
      if foot:
	 lines = lines + _indent(foot, indent_level)
      
      return lines


   def tags_str(self):
      """Return a string of all the tags."""

      s = []
      for tag, val in self.tags.items():
         s.append("%s=%s%s%s" % (tag, val[0], val[1], val[0]))
      return string.join(s)


   def set_tag(self, tag, val):
      """Set a tag on the element."""

      val = _escape(str(val))
      if '"' in val:
         if "'" in val:
            raise ValueError, "can't use tags containing both ' and \""
         else:
            quote_char = "'"
      else:
         quote_char = '"'
      self.tags[string.upper(tag)] = (quote_char, val)


   def set_default_tag(self, tag, val):
      """Sets the default value for a tag.

      This will only set a value if there is no existing tag by the
      given name.

      """

      if string.upper(tag) not in self.tags.keys():
         self.set_tag(tag, val)

   def get_head(self):
      """Returns the header for this element.

      Overload this to get some sort of begin element put on.  It should
      return a list of strings to be printed as the header.  Also take a
      look at element.get_indent()

      """

      if (self.head):
	 tags = self.tags_str()
	 if tags:
	    head = string.join([self.head, tags])
	 else:
	    head = self.head
	 return [self.prefix_string +
		 '<' + head + '>']
      elif self.prefix_string:
	 return [self.prefix_string]
      else:
	 return []


   def get_foot(self):
      """Returns the footer for this element.

      Overload this to get some sort of end element put on.  It should
      return a list of strings to be printed as the footer.  This should
      probably never be overloaded except for container elements."""

      if self.postfix_string:
	 return [self.postfix_string]
      else:
	 return []


   def get_indent(self):
      """Returns the indentation level for enclosed stuff.

      This is the number of indentation stops to go in for things inside
      this element.  Because only containers can actually contain
      things, this should definitely never be defined for container
      elements."""

      return self.indent_level


   def prefix(self, string):
      """Prefixes the element.

      This will tack the given string on to the *same line* as the
      element itself.  This function will always add to the beginning of
      the element, so if called twice, the second string will be put on
      before the first."""

      self.prefix_string = string + self.prefix_string
      return self


   def postfix(self, string):
      """Postfixes the element.

      This will tack the given string on to the *same line* as the
      element itself.  This function will always add to the end of the
      element, so if called twice, the second string will be put on
      after the first."""

      self.postfix_string = self.postfix_string + string
      return self


class container(element):
   """Base container class.

   This is an element that can contain things inside it.

   Set escape to 0 if you do *not* want the stuff inside escaped.  This
   might not work the way you think it should, because objects aren't
   evaluated until __repr__ is called.  Better to use the "literal"
   container class.

   Set newlines to 0 if you don't want a newline after the opening
   element and before the closing one.  That is, when you want:

   <ELEMENT>blah</ELEMENT>

   instead of:

   <ELEMENT>
     blah
   </ELEMENT>"""

   data = []
   inputs = []
   escape = 1
   newlines = 1
   indent_level = 1

   def __init__(self, *data, **tags):
      """Extract useful stuff."""

##       self.data = []
##       self.inputs = []
##       self.escape = 1
##       self.newlines = 1

      apply(element.__init__, (self,), tags)

      self.data = []
      if data:
	 apply(self.append, data)



   def __repr__(self):
      list = self.list()

      try:
	 return string.join(list, '\n')
      except TypeError:
	 raise TypeError, "all elements of .list() must be strings"


   def list(self, indent_level=0):
      out = []

      # First, adjust the indentation level
      contents_indent_level = indent_level + self.get_indent()
      
      for i in self.data:
	 typ = type(i)
	 if typ == types.StringType:
	    # String
	    lines = string.split(i, '\n')
	    if self.escape:
	       lines = map(lambda s: _escape(s),
			   lines)
	    out = out + _indent(lines, contents_indent_level)

         elif typ == types.ListType:
            # List
            lines = i
            if self.escape:
               lines = map(lambda s: _escape(s),
                           lines)
            out = out + _indent(lines, contents_indent_level)

	 elif typ == types.InstanceType:
	    # Element or Container
	    out = out + i.list(contents_indent_level)
	    try:
	       self.inputs = self.inputs + i.inputs
	    except AttributeError:
	       pass

	 elif typ == types.NoneType:
	    # Nothing
	    pass

	 else:
	    raise TypeError, 'invalid type in <%s>: %s (%s)' \
                  % (self.__class__.__name__, `i`, `typ`)


      # Find out how much the guts are indented
      indents = len(_indent([''], contents_indent_level)[0])

      # Prepend the header...
      head = self.get_head()
      if head:
	 head = _indent(head, indent_level)
	 if self.newlines:
	    out = head + out
	 else:
	    try:
	       out[0] = head[-1] + out[0][indents:]
	    except IndexError:
	       out = [head[-1]]
	    out = head[:-1] + out

      # Append the footer!
      foot = self.get_foot()
      if foot:
	 if self.newlines:
	    foot = _indent(foot, indent_level)
	    out = out + foot
	 else:
	    try:
	       out[-1] = out[-1] + foot[0]
	    except IndexError:
	       out = foot[0]
	    out = out + foot[1:]

      return out


   def get_foot(self):
      """Overload this to get a different end element put on.  By
      default, it uses the head element with a / in front of it (if
      head() returns a string)"""

      head = self.head
      if (head):
	 return ['</' + head + '>' + self.postfix_string]
      elif self.postfix_string:
	 return [self.postfix_string]
      else:
	 return []


   def prepend(self, *data):
      self.data = list(data) + self.data
      return self
      

   def append(self, *data):
      for item in data:
	 if isinstance(item, element):
	    self.data.append(item)
	 else:
	    self.data.append(str(item))
	    
      return self


class list_container(container):
   def __init__(self, *data, **tags):
      apply(container.__init__, (self,), tags)
      for i in data:
	 self.append(li(i))

   def item(self, *data, **tags):
      self.append(apply(li, data, tags))



class document(container):
   """An HTML document.  I went ahead and defined a par() and an h()
   method that will let you insert those things without having to
   instantiate your own class."""

   head = 'BODY'
   indent_level = 2

   def __init__(self, title, *data, **tags):
      apply(container.__init__, (self,) + (data), tags)
      self.title = title


   def get_head(self):
      l = []
      l.append('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2//EN">')
      l.append('<HTML>')
      l.append('  <HEAD>')
      l.append('    <TITLE>' + self.title + '</TITLE>')
      l.append('    <META NAME="generator" CONTENT="html.py">')
      l.append('  </HEAD>')
      l.append('  <' + string.join(['BODY'] + [self.tags_str()]) + '>')

      return l


   def get_foot(self):
      return ['  </BODY>', '</HTML>']


   def par(self, *data, **tags):
      self.append(apply(par, data, tags))

   def h(self, level, *data, **tags):
      self.append(apply(header, (level,) + data, tags))


class form(container):
   """An instance of an HTML form.  The default method is "post", unless
   you pass in your own method tag."""

   def __init__(self, action, *data, **tags):
      apply(container.__init__, (self,) + data, tags)
      self.set_default_tag('METHOD', 'post')
      self.set_default_tag('ACTION', action)


   def par(self, *data, **tags):
      self.append(apply(par, data, tags))


   def textarea(self, *data, **tags):
      self.append(apply(textarea, data, tags))
      

   def input(self, **tags):
      self.append(apply(input, (), tags))




class table(container):
   """A table."""

   def row(self, *data, **tags):
      row = tr()
      for i in (data):
	 row.append(apply(td, (i,), tags))
      self.append(row)


class h(container):
   """A header thingy."""

   def __init__(self, level, *data, **tags):
      self.head = 'H' + `level`
      apply(container.__init__, (self,) + data, tags)

def header(level, *data, **tags):
   """A header thingy."""
   
   return apply(h, (level,) + data, tags)


class escape(container):
   """Escaped text -- '&', '<', and '>' are replaced with SGML entities.

   This is essentially an empty container, since everything is escaped
   anyhow."""

   head = 0
   newlines = 0
   indent_level = 0


class literal(container):
   """Literal, unescaped, text.

   This class adds no header nor footer, and just prints out its
   contents verbatim.  Any containers inside this class *will* be
   escaped."""

   head = 0
   newlines = 0
   escape = 0
   indent_level = 0


class pre(container):
   """Returns preformatted text.

   The text inside will not be indented, but containers will be
   processed as usual.  That means their containers' data will be
   indented.  It would look something like this:

  <P>
    Here is some preformatted text:
  </P>
  <PRE>Blah blah blah
blah blah ...   blah
<ELEMENT>
  blah blah blah
  <ELEMENT2>
    blah blah blah
  </ELEMENT2>
</ELEMENT></PRE>
  <P>
    All done!
  </P>"""

   newlines = 0
   indent_level = -66536
#   def __init__(self, *data, **tags):
#      apply(container.__init__, (self,) + data, tags)
#      self.head = 0

##    def list(self, indentation_level=0):
##       lines = container.list(self, -1)
      
##       lines[0] = (indentation * indentation_level) + '<PRE>' + lines[0]
##       lines[-1] = lines[-1] + '</PRE>'

##       return lines


class a(container):
   """An anchor."""

   newlines = 0


def href(URI, *data, **tags):
   """A hyperlink."""

   tmp = apply(a, data, tags)
   tmp.set_default_tag( 'href', URI)
   return tmp

def link(URI, *data, **tags):
   return apply(href, (URI,) + data, tags)



class comment(container):
   """A comment.

   Although not truly a container, it makes more sense to do it this
   way."""

   newlines = 0

   def get_head(self):
      return ['<!-- ']
   
   def get_foot(self):
      return [' -->']


###
### Wussy classes
###

class p(container):
   """A paragraph."""

   pass

def par(*data, **tags):
   """A paragraph."""

   return apply(p, data, tags)


class tr(container):
   """A table row."""

   def td(self, *data, **tags):
      self.append(apply(td, data, tags))

   def th(self, *data, **tags):
      self.append(apply(th, data, tags))


class td(container):
   """A table cell."""

   def par(self, *data, **tags):
      self.append(apply(par, data, tags))


class th(container):
   """A table heading."""

   def par(self, *data, **tags):
      self.append(apply(par, data, tags))


class ul(list_container):
   """An unordered list."""

   pass

class ol(list_container):
   """An ordered (usually enumerated) list."""

   pass


class li(container):
   """A list item."""

   pass


class dl(container):
   """A dictionary list."""

   def title(self, *data, **tags):
      self.append(apply(dt, data, tags))

   def description(self, *data, **tags):
      self.append(apply(dd, data, tags))

   def pair(self, title, description):
       self.append(dt(title), dd(description))


class dt(container):
   """Dictionary list text part.

   This is the first of two parts to a dictionary list entry, and is
   usually a title or small amount of text which the second part
   describes in more detail."""

   pass

class dd(container):
   """Dictionary list definition (description) part.

   This is the second of two parts to a dictionary list entry, and is
   usually a longer description or explanation of the first part."""

   pass


class div(container):
   """A division.

   These are usually used to align or group chunks of HTML.  For
   instance, the preferred HTML 3.2 method of centering text is to use a
   DIV element with ALIGN="center".  To create one of these, use
   div(align='center')."""

   pass



class select(container):
   """A selection.

   Used as part of a form."""
   
   def option(self, *data, **tags):
      self.append(apply(option, data, tags))


class textarea(container):
   """A text area.

   Used as part of a form."""

   newlines = 0
   indent_level = -66536



class option(container):
   """An option.

   This *must* appear inside a select element."""
   
   pass


class samp(container):
   """Sample text.

   Use this to provide sample keyboard input.  Usually rendered in a
   monospaced font.

   """
   
   pass

class tt(container):
   """Typewriter text.

   Use this to provide the markup tag to specify a typewriter font.
   Please consider using logical markup such as <samp> instead, as this
   might not get your point across on speech synthesizers or
   non-graphical clients.

   """

   pass


class font(container):
   """Font modification.

   You can use this to change color or font size.  Please consider the
   use of small or big tags instead of this."""

   pass


class em(container):
   """Emphasized text.

   Use this to provide emphasis to something.  Usually rendered in
   italics.

   Please use this instead of an <i> element, unless you really must
   have italics for some reason.  Not all web browsers can do italics,
   but most will do something meaningful if given an emphasize element
   (for example, text browsers and browsers for the blind)."""
   
   pass


class strong(container):
   """Strong text.

   Use this to make something stand out.  Usually rendered in boldface.

   Please use this instead of a <b> element, unless you really must have
   boldface for some reason.  Not all web browsers can do boldface, but
   most will do something meaningful if given an strong element (for
   example, text browsers and browsers for the blind)."""

   pass


class i(container):
   """Italicized text.

   Consider using em if what you really want is emphasis, and not just
   italics."""

   pass


class b(container):
   """Boldface text.

   Consider using strong if what you really want is highlighted text,
   and not just boldface."""

   pass


class code(container):
   """Code sample.

   Use this to provide a sample of code.  Usually rendered in
   monospace."""
   
   pass


class small(container):
   """Small text.

   Please consider using this instead of a font element."""

   pass


class big(container):
    """Big text.

    Please consider using this instead of a font element."""

    pass


class blockquote(container):
    """Block quote.

    Generally rendered as indented on the left and right margin."""

    pass


##
## Elements
##

class input(element):
   """A form input."""
   
   def __init__(self, **tags):
      apply(element.__init__, (self,), tags)
      try:
	 self.inputs = [tags['name']]
      except KeyError:
	 pass

   def get_foot(self):
      return 0


class br(element):
   """A break.

   This causes the text to be broken, and all further text will appear
   on the next "line".  Do not confuse this with the <p> element--use
   this only for a break in text and *not* to start a new paragraph."""

   pass

class hr(element):
   """A horizontal rule."""
   
   pass


class img(element):
   """An image element.

   This element requires you to provide text for the image (the text
   will show up in an ALT element), since not all browsers can display
   images.

   <SOAPBOX>
     Think about why you are using this image.  Many images on
     web pages today add nothing to the content.  Here are some guidelines
     for what to set the text to:

     If the image is...

     text
       Set the text to be the text in the image.  Consider using an
       <h> element instead, since low-resolution displays are likely to
       make your image disproportionatley large, and high-resolution
       displays are likely to make your image unreadably small.
       
     decoration
       Set the text to be an empty string, since the image does not
       contribute to the page's content.

     photograph
       Set the text to a description of the photograph.  Wrap the
       image in a hyperlink to a larger size of the image, or to
       the image itself if no larger version is available.  This will
       make life much easier on Lynx users.
   </SOAPBOX>

   """
   
   def __init__(self, URI, text, **tags):
      apply(element.__init__, (self,), tags)
      self.set_default_tag('src', URI)
      self.set_default_tag('alt', text)


##
## Private (?) stuff
##

def _indent(lines, indentation_level):
   """Indents lines indentation_level times."""
   
   out = []
   indent = indentation * indentation_level
   return map(lambda s, indent=indent: indent + s, lines)


def _escape(s):
   """Escapes lines for HTML.

   Please don't use this.  Use the escape() class instead, it can behave
   just like a function."""
   
   # I think this split/join might be faster than the regsub method used
   # in cgi.py.

   s = string.join(string.split(s, '&'), '&amp;') # Must be done first!
   s = string.join(string.split(s, '<'), '&lt;')
   s = string.join(string.split(s, '>'), '&gt;')
   return s


def _set_default_tag(tags, name, value):
   """Set a default for a tag.

   What this does is to set a value if and only if it is not already
   set.  Pretty dumb, but it was something I was doing often enough to
   make it a function."""

   if (not tags.has_key(name)) and value:
      tags[name] = value

   return tags
